import 'source-map-support/register';
import * as crypto from 'crypto';
import decamelize from 'decamelize';
import * as inflection from 'inflection';
import Bonjour, { Service } from 'bonjour';
import { EventEmitter } from 'events';
import { get, put } from 'request-promise-native';

import { Services, Characteristics } from './hap-types';
import { HapMonitor } from './monitor';
import { IHapAccessoriesRespType, IServiceType, ICharacteristicType, IHapInstance, createDefaultCharacteristicType, IAccessoryResp } from './interfaces';

export type HapAccessoriesRespType = IHapAccessoriesRespType;
export type ServiceType = IServiceType;
export type CharacteristicType = ICharacteristicType;

export interface IDevice {
    addresses: Array<string>
    fqdn: string,
    host: string,
    name: string,
    port: number,
    protocol: "tcp" | "udp",
    rawTxt: Buffer,
    referer: {
        address: string,
        family: "IPv4" | "IPv6",
        port: number,
        size: number,
    },
    subtypes: Array<any>,
    txt: any,
    type: string
}

interface IConfig {
    debug?: boolean;
    instanceBlacklist?: string[];
}

export class HapClient {
    private bonjour = Bonjour();
    private browser?: Bonjour.Browser;
    private discoveryInProgress = false;

    private log: (msg: string) => void;
    private pin: string;
    private debugEnabled: boolean;
    private config: IConfig = {
        debug: true,
        instanceBlacklist: [],
    }

    private instances: IHapInstance[] = [];

    private hiddenServices = [
        Services.AccessoryInformation
    ];

    private hiddenCharacteristics = [
        Characteristics.Name
    ];

    constructor(opts: {
        pin: string;
        logger?: any;
        config: any;
    }) {
        this.pin = opts.pin;
        this.log = opts.logger;
        this.debugEnabled = true;
        this.config = opts.config;
    }

    private debug(msg: any) {
        if (this.debugEnabled) {
            this.log(msg);
        }
    }

    private debugErr(msg: string, err: Error) {
        if (this.debugEnabled) {
            this.log(`${msg}: ${err.message}\n ${err.stack}`);
        }
    }

    public refreshInstances() {
        if (!this.discoveryInProgress) {
            this.discover();
        } else {
            try {
                this.debug(`[HapClient] Discovery :: Re-broadcasting discovery query`);
                if (this.browser) {
                    this.browser.update();
                }

            } catch{ }
        }
    }

    public async discover(): Promise<void> {
        return new Promise((resolve) => {
            this.discoveryInProgress = true;

            this.browser = this.bonjour.find({
                type: 'hap'
            });

            // start matching services
            this.browser.start();
            this.debug(`[HapClient] Discovery :: Started`);

            // service found
            this.browser.on('up', async (service: any) => {
                let device = service as IDevice;
                if (!device || !device.txt) {
                    this.debug(`[HapClient] Discovery :: Ignoring device that contains no txt records. ${JSON.stringify(device)}`);
                    return;
                }

                const instance: IHapInstance = {
                    displayName: device.name,
                    ipAddress: device.addresses[0],
                    name: device.txt.md,
                    username: device.txt.id,
                    port: device.port,
                }

                this.debug(`[HapClient] Discovery :: Found HAP device ${instance.displayName} with username ${instance.username}`);

                // update an existing instance
                const existingInstanceIndex = this.instances.findIndex(x => x.username === instance.username);
                if (existingInstanceIndex > -1) {

                    if (
                        this.instances[existingInstanceIndex].port !== instance.port ||
                        this.instances[existingInstanceIndex].name !== instance.name
                    ) {
                        this.instances[existingInstanceIndex].port = instance.port;
                        this.instances[existingInstanceIndex].name = instance.name;
                        this.debug(`[HapClient] Discovery :: [${this.instances[existingInstanceIndex].ipAddress}:${instance.port} ` +
                            `(${instance.username})] Instance Updated`);
                    }

                    return;
                }

                //Comenting out because of lack of config

                //check instance is not on the blacklist
                if (this.config.instanceBlacklist && this.config.instanceBlacklist.find(x => instance.username.toLowerCase() === x.toLowerCase())) {
                    this.debug(`[HapClient] Discovery :: Instance with username ${instance.username} found in blacklist. Disregarding.`);
                    return;
                }

                for (const ip of device.addresses) {
                    if (ip.match(/^(?:(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])(\.(?!$)|$)){4}$/)) {
                        try {
                            this.debug(`[HapClient] Discovery :: Testing ${instance.displayName} ${instance.username} via http://${ip}:${device.port}/accessories`);
                            const test = await get(`http://${ip}:${device.port}/accessories`, {
                                json: true,
                                timeout: 1000,
                                headers: { Authorization: this.pin }
                            });
                            if (test.accessories) {
                                this.debug(`[HapClient] Discovery :: Success ${instance.displayName} ${instance.username} via http://${ip}:${device.port}/accessories`);
                                instance.ipAddress = ip;
                            }
                            break;
                        } catch (e) {
                            this.debug(`[HapClient] Discovery :: ***Failed*** ${instance.displayName} ${instance.username} via http://${ip}:${device.port}/accessories: ${e.message}`);
                        }
                    }
                }

                // store instance record
                if (instance.ipAddress) {
                    this.instances.push(instance);
                    this.debug(`[HapClient] Discovery :: [${instance.displayName} - ${instance.ipAddress}:${instance.port} (${instance.username})] Instance Registered`);
                } else {
                    this.debug(`[HapClient] Discovery :: Could not register to device ${instance.displayName} with username ${instance.username}`);
                }
            });

            // stop discovery after 20 seconds
            setTimeout(() => {
                if (this.browser) {
                    this.browser.stop();
                }
                this.debug(`[HapClient] Discovery :: Ended`);
                this.discoveryInProgress = false;
                return resolve();
            }, 20000);
        });
    }

    private humanizeString(string: string) {
        return inflection.titleize(decamelize(string));
    }

    public async getAccessories(): Promise<Array<IAccessoryResp>> {
        if (!this.instances.length) {
            this.debug('[HapClient] Cannot load accessories. No Homebridge instances have been discovered.');
        }

        const accessories: Array<IAccessoryResp> = [];
        for (const instance of this.instances) {
            try {
                const resp: IHapAccessoriesRespType = await get(`http://${instance.ipAddress}:${instance.port}/accessories`, { json: true });
                resp.accessories && resp.accessories.forEach((accessory: IAccessoryResp) => {
                    accessory.instance = instance;
                    accessories.push(accessory);
                })
            } catch (e) {
                if (this.log) {
                    this.debugErr(`[HapClient] [${instance.displayName} - ${instance.ipAddress}:${instance.port} (${instance.username})] Failed to connect`, e);
                }
            }
        }
        return accessories;
    }

    public async monitorCharacteristics(): Promise<HapMonitor> {
        const services = await this.getAllServices();
        return new HapMonitor(this.log, this.debug.bind(this), this.pin, services);
    }

    public async getAllServices(): Promise<Array<IServiceType>> {
        /* Get Accessories from HAP */
        const accessories = await this.getAccessories();

        const services: Array<IServiceType> = [];

        /* Parse All Accessories */
        accessories.forEach(accessory => {

            /* Parse Accessory Information */
            const accessoryInformationService = accessory.services.find(x => x.type === Services.AccessoryInformation);
            const accessoryInformation: { [key: string]: any } = {};

            if (accessoryInformationService && accessoryInformationService.characteristics) {
                accessoryInformationService.characteristics.forEach((c) => {
                    if (c.value) {
                        accessoryInformation[c.description] = c.value;
                    }
                });
            }

            /* Parse All Services */
            accessory.services
                .filter((s) => this.hiddenServices.indexOf(s.type) < 0 && Services[s.type])
                .map((s) => {
                    let serviceName = s.characteristics.find(x => x.type === Characteristics.Name);

                    /* Set default name characteristic if none defined */
                    serviceName = serviceName ? serviceName : {
                        iid: 0,
                        type: Characteristics.Name,
                        description: 'Name',
                        format: 'string',
                        value: this.humanizeString(Services[s.type]),
                        perms: ['pr']
                    };

                    /* Parse Service Characteristics */
                    const serviceCharacteristics: Array<ICharacteristicType> = s.characteristics
                        .filter((c) => this.hiddenCharacteristics.indexOf(c.type) < 0 && Characteristics[c.type])
                        .map((c) => {
                            return {
                                aid: accessory.aid,
                                iid: c.iid,
                                uuid: c.type,
                                type: Characteristics[c.type],
                                serviceType: Services[s.type],
                                serviceName: serviceName!.value.toString(),
                                description: c.description,
                                value: c.value,
                                format: c.format,
                                perms: c.perms,
                                unit: c.unit,
                                maxValue: c.maxValue,
                                minValue: c.minValue,
                                minStep: c.minStep,
                                canRead: c.perms.includes('pr'),
                                canWrite: c.perms.includes('pw'),
                                ev: c.perms.includes('ev'),
                            };
                        });

                    const service: IServiceType = {
                        aid: accessory.aid,
                        iid: s.iid,
                        uuid: s.type,
                        type: Services[s.type],
                        humanType: this.humanizeString(Services[s.type]),
                        serviceName: serviceName.value.toString(),
                        serviceCharacteristics: serviceCharacteristics,
                        accessoryInformation: accessoryInformation,
                        values: {},
                        linked: s.linked,
                        instance: accessory.instance,
                    };

                    // generate unique id for service
                    service.uniqueId = crypto.createHash('sha256')
                        .update(`${service.instance.username}${service.aid}${service.iid}${service.type}`)
                        .digest('hex');

                    /* Helper function to trigger a call to the accessory to get all the characteristic values */
                    service.refreshCharacteristics = () => {
                        return this.refreshServiceCharacteristics.bind(this)(service);
                    };

                    /* Helper function to set the value of a characteristic */
                    service.setCharacteristic = (iid: number, value: number | string | boolean): Promise<IServiceType | ICharacteristicType> => {
                        return this.setCharacteristic.bind(this)(service, iid, value);
                    };

                    /* Helper function to returns a characteristic by it's type name */
                    service.getCharacteristic = (type: string): ICharacteristicType => {
                        let characteristic = service.serviceCharacteristics.find(c => c.type === type);
                        return characteristic ? characteristic : createDefaultCharacteristicType();
                    };

                    service.serviceCharacteristics.forEach((cType: ICharacteristicType) => {
                        /* Helper function to set the value of a characteristic */
                        cType.setValue = async (value: number | string | boolean): Promise<ICharacteristicType> => {
                            return await this.setCharacteristic.bind(this)(service, cType.iid, value);
                        };

                        /* Helper function to get the value of a characteristic from the accessory */
                        cType.getValue = async (): Promise<ICharacteristicType> => {
                            return await this.getCharacteristic.bind(this)(service, cType.iid);
                        };

                        /* set the values for each characteristic type in an easy-to-access object */
                        service.values[cType.type] = cType.value;
                    });

                    services.push(service);
                });
        });

        return services;
    }

    public getService(iid: number): Promise<IServiceType | undefined> {
        return new Promise(async (resolve, reject) => {
            try {
                const services = await this.getAllServices();
                return resolve(services.find(x => x.iid === iid));
            } catch (err) {
                return reject(err);
            }
        });

    }

    public async getServiceByName(serviceName: string): Promise<IServiceType | undefined> {
        return new Promise(async (resolve, reject) => {
            try {
                const services = await this.getAllServices();
                return resolve(services.find(x => x.serviceName === serviceName));
            } catch (err) {
                return reject(err);
            }
        });
    }

    public async refreshServiceCharacteristics(service: IServiceType): Promise<IServiceType> {
        const iids: number[] = service.serviceCharacteristics.map(c => c.iid);

        const resp = await get(`http://${service.instance.ipAddress}:${service.instance.port}/characteristics`, {
            qs: {
                id: iids.map(iid => `${service.aid}.${iid}`).join(',')
            },
            json: true
        });

        resp.characteristics.forEach((charType: ICharacteristicType) => {
            const characteristic = service.serviceCharacteristics.find(x => x.iid === charType.iid && x.aid === service.aid);
            characteristic!.value = charType.value;
        });

        return service;
    }

    public async getCharacteristic(service: IServiceType, iid: number): Promise<ICharacteristicType> {
        const resp = await get(`http://${service.instance.ipAddress}:${service.instance.port}/characteristics`, {
            qs: {
                id: `${service.aid}.${iid}`
            },
            json: true
        });

        const characteristic = service.serviceCharacteristics.find(x => x.iid === resp.characteristics[0].iid && x.aid === service.aid);
        if (characteristic) {
            characteristic.value = resp.characteristics[0].value;
        }

        return characteristic ? characteristic : createDefaultCharacteristicType();
    }

    public async setCharacteristic(service: IServiceType, iid: number, value: number | string | boolean): Promise<ICharacteristicType> {
        let characteristic = createDefaultCharacteristicType();
        try {
            await put(`http://${service.instance.ipAddress}:${service.instance.port}/characteristics`, {
                headers: {
                    Authorization: this.pin
                },
                json: {
                    characteristics: [
                        {
                            aid: service.aid,
                            iid: iid,
                            value: value
                        }
                    ]
                }
            });

            characteristic = await this.getCharacteristic(service, iid);
        } catch (e) {
            if (this.log) {
                this.debugErr(`[HapClient] [${service.instance.ipAddress}:${service.instance.port} (${service.instance.username})] ` +
                    `Failed to set value for ${service.serviceName}.`, e);
                if (e.statusCode === 401) {
                    this.debug(`[HapClient] [${service.instance.ipAddress}:${service.instance.port} (${service.instance.username})] ` +
                        `Make sure Homebridge pin for this instance is set to ${this.pin}.`);
                } else {
                    this.debug(e.message);
                }
            } else {
                console.log(e);
            }
        }

        return characteristic
    }
}
